import pytest

from cellseg_models_pytorch.models import cellpose_base
from cellseg_models_pytorch.optimizers import OPTIM_LOOKUP, adjust_optim_params


@pytest.mark.parametrize("optim", ["adam"])
def test_optim_setup(optim):
    model = cellpose_base(type_classes=3)
    params = adjust_optim_params(
        model,
        encoder_lr=0.5,
        encoder_wd=0.4,
        decoder_lr=0.3,
        decoder_wd=0.2,
        remove_bias_wd=True,
    )

    optimizer = OPTIM_LOOKUP[optim](params)

    assert all([p_g["lr"] == p["lr"] for p_g, p in zip(optimizer.param_groups, params)])
    assert all(
        [
            p_g["weight_decay"] == p["weight_decay"]
            for p_g, p in zip(optimizer.param_groups, params)
        ]
    )



import torch
import torch.nn as nn

from ..weighted_base_loss import WeightedBaseLoss


class CELoss(WeightedBaseLoss):
    def __init__(
        self, edge_weight: float = None, class_weights: torch.Tensor = None, **kwargs
    ) -> None:
        """Cross-Entropy loss with weighting.

        Parameters
        ----------
            edge_weight : float, default=none
                Weight that is added to object borders.
            class_weights : torch.Tensor, default=None
                Class weights. A tensor of shape (n_classes,).
        """
        super().__init__(class_weights, edge_weight)
        self.loss = nn.CrossEntropyLoss(reduction="none", weight=class_weights)

    def forward(
        self,
        yhat: torch.Tensor,
        target: torch.Tensor,
        target_weight: torch.Tensor = None,
        **kwargs
    ) -> torch.Tensor:
        """Compute the cross entropy loss.

        Parameters
        ----------
            yhat : torch.Tensor
                The prediction map. Shape (B, C, H, W).
            target : torch.Tensor
                the ground truth annotations. Shape (B, H, W).
            target_weight : torch.Tensor, default=None
                The edge weight map. Shape (B, H, W).

        Returns
        -------
            torch.Tensor:
                Computed CE loss (scalar).
        """
        loss = self.loss(yhat, target)  # (B, H, W)

        if self.edge_weight is not None:
            loss = self.apply_edge_weights(loss, target_weight)

        return loss.mean()



from typing import Dict, List

import torch.nn as nn


def adjust_optim_params(
    model: nn.Module,
    encoder_lr: float,
    encoder_wd: float,
    decoder_lr: float,
    decoder_wd: float,
    remove_bias_wd: bool = True,
) -> List[Dict[str, Dict]]:
    """Adjust the learning parameters for optimizer.

    1. Adjust learning rate and weight decay in the pre-trained
        encoder and decoders.
    2. Remove weight decay from bias terms to reduce overfitting.

    "Bag of Tricks for Image Classification with Convolutional Neural Networks"
    - https://arxiv.org/pdf/1812.01187

    Parameters
    ----------
        model : nn.Module
            The encoder-decoder segmentation model.
        encoder_lr : float
            Learning rate of the model encoder.
        encoder_wd : float
            Weight decay for the model encoder.
        decoder_lr : float
            Learning rate of the model decoder.
        decoder_wd : float
            Weight decay for the model decoder.
        remove_bias_wd : bool, default=True
            If True, the weight decay from the bias terms is removed from the model
            params. Ignored if `remove_wd`=True.

    Returns
    -------
        List[Dict[str, Dict]]:
            a list of kwargs (str, Dict pairs) containing the model params.
    """
    params = list(model.named_parameters())
    encoder_params = {"encoder": {"lr": encoder_lr, "weight_decay": encoder_wd}}
    decoder_params = {"decoder": {"lr": decoder_lr, "weight_decay": decoder_wd}}

    adjust_params = []
    for name, parameters in params:
        opts = {}
        for enc, enc_opts in encoder_params.items():
            if enc in name:
                for key, item in enc_opts.items():
                    opts[key] = item

        for dec, dec_opts in decoder_params.items():
            if dec in name:
                for key, item in dec_opts.items():
                    opts[key] = item

        if remove_bias_wd:
            if name.endswith("bias"):
                opts["weight_decay"] = 0.0

        adjust_params.append({"params": parameters, **opts})

    return adjust_params



import torch
import torch.nn.functional as F

from cellseg_models_pytorch.utils import tensor_one_hot

from ..weighted_base_loss import WeightedBaseLoss


class MSE(WeightedBaseLoss):
    def __init__(
        self, edge_weight: float = None, class_weights: torch.Tensor = None, **kwargs
    ) -> None:
        """MSE-loss.

        Parameters
        ----------
            edge_weight : float, default=none
                Weight that is added to object borders.
            class_weights : torch.Tensor, default=None
                Class weights. A tensor of shape (n_classes,).
        """
        super().__init__(class_weights, edge_weight)

    def forward(
        self,
        yhat: torch.Tensor,
        target: torch.Tensor,
        target_weight: torch.Tensor = None,
        **kwargs
    ) -> torch.Tensor:
        """Compute the MSE-loss.

        Parameters
        ----------
            yhat : torch.Tensor
                The prediction map. Shape (B, C, H, W).
            target : torch.Tensor
                the ground truth annotations. Shape (B, H, W).
            target_weight : torch.Tensor, default=None
                The edge weight map. Shape (B, H, W).

        Returns
        -------
            torch.Tensor:
                Computed MSE loss (scalar).
        """
        target_one_hot = target
        if target.size() != yhat.size():
            if target.dtype == torch.float32:
                target_one_hot = target.unsqueeze(1)
            else:
                target_one_hot = tensor_one_hot(target, yhat.shape[1])

        mse = F.mse_loss(yhat, target_one_hot, reduction="none")  # (B, C, H, W)
        mse = torch.mean(mse, dim=1)  # to (B, H, W)

        if self.class_weights is not None:
            mse = self.apply_class_weights(mse, target)

        if self.edge_weight is not None:
            mse = self.apply_edge_weights(mse, target_weight)

        return torch.mean(mse)



import torch
import torch.nn as nn


class WeightedBaseLoss(nn.Module):
    def __init__(
        self, class_weights: torch.Tensor = None, edge_weight: float = None
    ) -> None:
        """Init a base class for weighted cross entropy based losses.

        Enables weighting for object instance edges and classes.

        Parameters
        ----------
            class_weights : torch.Tensor, default=None
                Class weights. A tensor of shape (C, )
            edge_weight : float, default=None
                Weight for the object instance border pixels
        """
        super().__init__()
        self.class_weights = class_weights
        self.edge_weight = edge_weight

    def apply_class_weights(
        self, loss_matrix: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """Multiply pixelwise loss matrix by the class weights.

        Note: No normalization

        Parameters
        ----------
            loss_matrix : torch.Tensor
                Pixelwise losses. A tensor of shape (B, H, W).
            target : torch.Tensor
                The target mask. Shape (B, H, W).

        Returns
        -------
            torch.Tensor:
                The loss matrix scaled with the weight matrix. Shape (B, H, W).
        """
        weight_mat = self.class_weights[target].to(target.device)  # to (B, H, W)
        loss = loss_matrix * weight_mat

        return loss

    def apply_edge_weights(
        self, loss_matrix: torch.Tensor, weight_map: torch.Tensor
    ) -> torch.Tensor:
        """Apply weights to the object boundaries.

        Basically just computes `edge_weight`**`weight_map`.

        Parameters
        ----------
            loss_matrix : torch.Tensor
                Pixelwise losses. A tensor of shape (B, H, W).
            weight_map : torch.Tensor
                Map that points to the pixels that will be weighted.
                Shape (B, H, W).

        Returns
        -------
            torch.Tensor:
                The loss matrix scaled with the nuclear boundary weights.
                Shape (B, H, W).
        """
        return loss_matrix * self.edge_weight**weight_map

    def extra_repr(self) -> str:
        """Add info to print."""
        s = "class_weights={class_weights}, edge_weight={edge_weight}"
        return s.format(**self.__dict__)


# def cross_entropy_with_label_smoothing(pred, target, label_smoothing=0.1):
#     logsoftmax = nn.LogSoftmax()
#     n_classes = pred.size(1)
#     # convert to one-hot
#     target = torch.unsqueeze(target, 1)
#     soft_target = torch.zeros_like(pred)
#     soft_target.scatter_(1, target, 1)
#     # label smoothing
#     soft_target = soft_target * (1 - label_smoothing) + label_smoothing / n_classes
#     return torch.mean(torch.sum(- soft_target * logsoftmax(pred), 1))



from typing import Dict, List

import numpy as np
from pathos.multiprocessing import ThreadPool as Pool
from skimage.filters import rank
from skimage.morphology import closing, disk, opening
from skimage.util import img_as_ubyte
from tqdm import tqdm

from ..postproc import POSTPROC_LOOKUP
from ..utils import binarize, fill_holes_semantic, remove_debris_semantic

__all__ = ["PostProcessor"]


INST_ARGMAX = {
    "stardist": False,
    "stardist_orig": False,
    "cellpose": True,
    "cellpose_old": True,
    "omnipose": True,
    "hovernet": True,
    "drfns": True,
    "dcan": False,
    "dran": False,
}


class PostProcessor:
    def __init__(
        self, instance_postproc: str, inst_key: str, aux_key: str, **kwargs
    ) -> None:
        """Multi-threaded post-processing.

        Parameters
        ----------
            instance_postproc : str
                The post-processing method for the instance segmentation mask.
            inst_key : str
                The key/name of the model instance prediction output that is used
                for the instance segmentation post-processing pipeline.
            aux_key : Tuple[str, ...]:
                The key/name of the model auxilliary output that is used for the
                instance segmentation post-processing pipeline.
            **kwargs
                Arbitrary keyword arguments that can be used for any of the private
                post-processing functions of this class.
        """
        allowed = list(POSTPROC_LOOKUP.keys())
        if instance_postproc not in allowed:
            raise ValueError(
                f"Illegal post proc method. Got {instance_postproc}. Allowed: {allowed}"
            )

        self.postproc_method = POSTPROC_LOOKUP[instance_postproc]
        self.use_argmax = INST_ARGMAX[instance_postproc]
        self.inst_key = inst_key
        self.aux_key = aux_key
        self.kwargs = kwargs

    def _get_sem_map(
        self,
        prob_map: np.ndarray,
        use_blur: bool = False,
        use_closing: bool = False,
        use_opening: bool = True,
        disk_size: int = 10,
        **kwargs,
    ) -> np.ndarray:
        """Run the semantic segmentation post-processing."""
        # Median filtering to get rid of noise. Adds a lot of overhead sop optional.
        if use_blur:
            sem = np.zeros_like(prob_map)
            for i in range(prob_map.shape[0]):
                sem[i, ...] = rank.median(
                    img_as_ubyte(prob_map[i, ...]), footprint=disk(disk_size)
                )
            prob_map = sem

        sem = np.argmax(prob_map, axis=0)

        if use_opening:
            sem = opening(sem, disk(disk_size))

        if use_closing:
            sem = closing(sem, disk(disk_size))

        sem = remove_debris_semantic(sem)
        sem = fill_holes_semantic(sem)

        return sem

    def _get_inst_map(
        self, prob_map: np.ndarray, aux_map: np.ndarray, **kwargs
    ) -> np.ndarray:
        """Run the instance labelling post processing."""
        if self.use_argmax:
            inst = np.argmax(prob_map, axis=0)
        else:
            inst = prob_map.squeeze()

        return self.postproc_method(inst, aux_map, **self.kwargs)

    def _get_type_map(
        self,
        prob_map: np.ndarray,
        inst_map: np.ndarray,
        use_mask: bool = False,
        **kwargs,
    ) -> np.ndarray:
        """Run the type map post-processing. Majority voting for each instance.

        Adapted from:
        https://github.com/vqdang/hover_net/blob/master/models/hovernet/post_proc.py
        """
        type_map = np.argmax(prob_map, axis=0)
        if use_mask:
            type_map = binarize(inst_map) * type_map

        pred_id_list = np.unique(inst_map)[1:]
        for inst_id in pred_id_list:
            inst_type = type_map[inst_map == inst_id]
            type_list, type_pixels = np.unique(inst_type, return_counts=True)
            type_list = list(zip(type_list, type_pixels))
            type_list = sorted(type_list, key=lambda x: x[1], reverse=True)
            cell_type = type_list[0][0]

            if cell_type == 0:
                if len(type_list) > 1:
                    cell_type = type_list[1][0]

            type_map[inst_map == inst_id] = cell_type

        return type_map

    def post_proc_pipeline(
        self, maps: Dict[str, np.ndarray], **kwargs
    ) -> Dict[str, np.ndarray]:
        """Run post-processing for all the model outputs.

        Parameters
        ----------
            maps : Dict[str, np.ndarray]
                model head names mapped to model outputs.
                E.g. {"sem": np.ndarray, "type": np.ndarray, "cellpose": np.ndarray}

        Returns
        -------
            Dict[str, np.ndarray]:
                final output names mapped to the final post-processed outputs.
                E.g. {"sem": np.ndarray, "type": np.ndarray, "inst": np.ndarray}
        """
        res = {}
        if "sem" in maps.keys():
            res["sem"] = self._get_sem_map(maps["sem"], **self.kwargs)

        res["inst"] = self._get_inst_map(
            maps[self.inst_key], maps[self.aux_key], **self.kwargs
        )

        if "type" in maps.keys():
            res["type"] = self._get_type_map(maps["type"], res["inst"], **self.kwargs)

        return res

    def run_parallel(
        self, maps: List[Dict[str, np.ndarray]], progress_bar: bool = False
    ) -> List[Dict[str, np.ndarray]]:
        """Run the full post-processing pipeline in parallel for many model outputs.

        Parameters
        ----------
            maps : List[Dict[str, np.ndarray]]
                The model output map dictionaries in a list.
            progress_bar : bool, default=False
                If True, a tqdm progress bar is shown.


        Returns
        -------
            List[Dict[str, np.ndarray]]:
                The post-processed output map dictionaries in a list.
        """
        seg_results = []
        with Pool() as pool:
            if progress_bar:
                it = tqdm(pool.imap(self.post_proc_pipeline, maps), total=len(maps))
            else:
                it = pool.imap(self.post_proc_pipeline, maps)

            for x in it:
                seg_results.append(x)

        return seg_results
